/* program to listen to tecan plate reader <-> injector communications
and switch on lights when injection is happening

Author: Florian Richter, Moeglich Lab, HU Berlin
florian.richter.1@hu-berlin.de

general comments:
to make stuff work, we have to connect pins
and change the #define SPICS line in the mcp_can_dfs.h file
as described in http://www.arduinoforum.de/arduino-Thread-Seeedstudio-CAN-Bus-Shield-auf-Mega-2560-R3
In addition, we have to connect pin2 of the shield to pin 2
of the mega board (for the interrupt to work), and also connect
pin9 of the shield to pin9 of the board (for unknown reasons)
*/

#include <mcp_can.h>
#include <SPI.h>
#include <LiquidCrystal.h>

unsigned char Flag_Recv = 0;
unsigned char len = 0;
unsigned char buf[9];
INT32U last_can_id = 0;

int button_pin = 5;

int led_pins[3] = {3,4,6}; //3 is led 1, 4 is test led, 6 is led 2
int num_led_pins = 3;
int light_intensity = 255;

int int_pot_pin = 0;
int aread_value = 1023;

LiquidCrystal lcd(28, 29, 26, 27, 24, 25, 22);

//unsigned long tot_num_can_msg = 0, tot_loop_calls = 0;

//define this as an int to be faster, overflow won't be problematic
//because we only care about differences 
unsigned int loop_calls = 0;
boolean report_serial = 0;     //set this to 1 to copy CAN messages to the Serial line
boolean transmit_can_messages = 0;

//the injection speed received from the plate reader.
//33 corresponds to the highest possible value (300ul/s)
int inject_speed = 300; 

//function to determine which leds to turn on
//based on the last digit of the volume value
bool
led_on_last_digit(
 int last_digit,
 int led_num )
{
  if( last_digit == 0 ){ return true; } //all channels on
  if( led_num == 4 ){ return true; } //test is always on

  if( last_digit == 1 ){
     if( led_num == 3 ){ return true; }
     else return false;
  } 

  if( last_digit == 2 ){
    if( led_num == 6 ) { return true; }
    else return false;
  }
  return true; 
}

unsigned long time1 =0, time2 =0, time_last_can_msg = 0;
//String msg_converted = "";

//helper function to get the user provided value
//from the machine value
int inject_speed_converter( int & input ){
  return (1.007 / input ) * 10306;
}

// 1281 for injector A, 1289 for injector B
int can_bus_pump_id = 1289;
float minimum_light_flash = 0.5; //injection signals that are shorter than 0.5s will be ignored

void set_tecan_can_filters()
{
 //CAN.init_Filt(0, 0, 262 );
 //CAN.init_Filt(1 ,0, 1286); 
}

void setup()
{
  Serial.begin(115200);
  aread_value = analogRead( int_pot_pin );
  light_intensity = map( aread_value, 0, 1023, 0, 255 );
  lcd.begin(16,2);
  lcd.setCursor(0,0);
  lcd.print("Wazzup yo?");
  delay(1000);
  lcd.setCursor(0,0);
  lcd.print("Intensity: "); lcd.print(light_intensity);
  lcd.setCursor(0,1);
  lcd.print("CAN init... ");
  time1 = micros();
  if( CAN.begin(CAN_100KBPS)==CAN_OK ){
    time2 = micros();
    lcd.print("OK");
    Serial.print("Can initialization took "); Serial.print( time2 - time1 ); Serial.println(" microseconds");
    Serial.println("can init ok");  // init can bus : baudrate = 500k
  }
  else{
    lcd.print("FAIL");
    Serial.println("can init didn't work");
  }
  set_tecan_can_filters();
  attachInterrupt(0, MCP2515_ISR, FALLING);     // start interrupt
  loop_calls = 0;
  for(int led = 0; led < num_led_pins; ++led) pinMode( led_pins[led], OUTPUT);
  time_last_can_msg = 0;
  //digitalWrite( led_pin, HIGH);
  //delay(1000);
  //digitalWrite(led_pin, LOW);
}

void MCP2515_ISR()
{
    Flag_Recv = 1;
}

void reply_to_can_message(
 unsigned char const msgbuf[],
 unsigned char const & length,
 INT32U const & msg_id
)
{
  if( !transmit_can_messages ) return;
  
  if( msg_id == 262 ){
   if( (length == 2 ) && (msgbuf[0] == 49 ) && (msgbuf[1] == 51 ) ){
     //13 from 262
     unsigned char reply[2] = { 32, 32 };
     CAN.sendMsgBuf( 1154, 0, 2, reply);
     if( report_serial ) Serial.println("TX1");
     return;
   }
   else if( (length == 3 ) && (msgbuf[0] == 96 ) && (msgbuf[1] == 0 ) && (msgbuf[2] == 49 ) ){
      //` |N| 1 from 262
      unsigned char reply[2] = { 49, 51 };
      CAN.sendMsgBuf( 1286, 0, 2, reply);
      if( report_serial ) Serial.println("TX2");
      return;
   }
  } //if msg_id == 262
  else if( msg_id == 128 ){
    //Serial.print("blp");
   if(   (length == 3 ) && (msgbuf[0] == 74 ) && (msgbuf[1] == 49 ) && (msgbuf[2] == 82 ) ){
      //J1R command from 128
      unsigned char reply0[0];
      CAN.sendMsgBuf( 257, 0, 0, reply0);
      delay(100);
      
      unsigned char reply1[2] = {96,0 }; //`|N|
      CAN.sendMsgBuf( 1281, 0, 2, reply1);
      delay(100);
      
      unsigned char reply2[2] = {49,51}; //13
      CAN.sendMsgBuf( 1281, 0, 2, reply2);
      
      if( report_serial ) Serial.println("TX3");
      return;    
   }
   else if( (length == 2 ) && (msgbuf[0] == 49 ) && (msgbuf[1] == 51 ) ){
      //13 from 128
      unsigned char reply[3] = {96,0, 49 }; //`|N|
      CAN.sendMsgBuf( 262, 0, 3, reply);
      if( report_serial ) Serial.println("TX5"); 
   }
  } //else if msg_id == 128
  else if( msg_id == 1281 ){
    if( (length == 2 ) && (msgbuf[0] == 49 ) && (msgbuf[1] == 51 ) ){
      //13 from 1281
      unsigned char reply[3] = {96,0, 49 }; //`|N|
      //CAN.sendMsgBuf( 262, 0, 3, reply);
      //if( report_serial ) Serial.println("TX4");
    }
  }//else if( msg_id == 1281 ){
}


void loop()
{
    loop_calls += 1;
    //INT8U rec_check = CAN.checkReceive();
    //INT8U msg_avail = CAN.CAN_MSGAVAIL;
    if( Flag_Recv == 1 ||  CAN.checkReceive() == 3  )    // check if get data
    {
      time_last_can_msg = millis();
      Flag_Recv = 0;       // clear flag
      //tot_num_can_msg += 1;
      last_can_id = CAN.getCanId();
      CAN.readMsgBuf(&len, buf);            // read data,  len: data length, buf: data buf
      if( report_serial ){
        Serial.print("CC"); //Serial.print( tot_num_can_msg); 
        Serial.print(loop_calls);
        Serial.print(" L"); Serial.print(len);
        Serial.print(" ID"); Serial.println(last_can_id);
        
        for(int i = 0; i<len; i++)            // print the data
        {
          //Serial.print( char(buf[i]) ); Serial.print(" ");
            if( buf[i] == 0 ){
              Serial.print("|N| "); 
              //  msg_converted = msg_converted + " NULL ";
            }
            else { Serial.print( char(buf[i]) ); Serial.print(" ");}      
        } //loop over gotten array
        Serial.println();
      } //if report_serial

      //check if this is message sets the injection speed
      //integers correspond to ascii values
      if( (last_can_id == can_bus_pump_id) && (buf[1] == 83 ) ){
        int inj_speed = 33;
  
        if( buf[4] > 57){
           inj_speed = (10 * (buf[2] -48)) + (buf[3]-48);
        }
        else if( buf[5] > 57 ){
           inj_speed = (100 * (buf[2]-48) ) + (10 *(buf[3]-48)) + (buf[4]-48);
        }

        else{ 
          Serial.print( "IJS error ");
          for(int ie = 0; ie<len; ie++) Serial.print( char(buf[ie] ));
          Serial.println();
        }
        inject_speed = inject_speed_converter( inj_speed);
        if(report_serial){ Serial.print("IJS set "); Serial.println( inject_speed); }
      } // injection speed reading over, if( (last_can_id == 1281) && (buf[1] == 83 ) )
      
      //now check if we have a dispense command
      else if( (last_can_id == can_bus_pump_id) && (buf[0] == 68 ) ){
        if( buf[len-1] != 82 ) Serial.println("D er1");
        int user_volume = 0;
        int digit_counter = 1;
        int last_digit = buf[len-2] - 48;
        for( int voldigit = len - 2; voldigit > 0; voldigit--){
           user_volume += ((buf[voldigit] - 48) * digit_counter);
           digit_counter*=10;
        }
        //Serial.print("buf[1] is "); Serial.print( buf[1]); Serial.print(" in int is "); Serial.println(buf[1]-48); 
        float light_seconds = float(user_volume) / inject_speed;
        if(  light_seconds >= minimum_light_flash ) {
          //Serial.print("UVol "); Serial.print( user_volume);
          //Serial.print("ISp "); Serial.print( inject_speed);
          if(report_serial){ Serial.print("Light for "); Serial.print(light_seconds); Serial.println(" s");}
          
          for(int led = 0; led < num_led_pins; ++led) {
            if( led_on_last_digit( last_digit, led_pins[led] ) ){
              //Serial.print("writing pin "); Serial.println( led_pins[led] );
              analogWrite( led_pins[led], light_intensity);
            }
          }
          
          lcd.clear(); lcd.setCursor(0,0); lcd.print("Light on for "); lcd.setCursor(0,1);
          lcd.print(light_seconds); lcd.print("s");
          delay( 1000* light_seconds );
          for(int led = 0; led < num_led_pins; ++led) digitalWrite( led_pins[led], LOW);
          lcd.clear();
        }
      } //dispense command checking over, if( (last_can_id == 1281) && (buf[0] == 68 )
      
     reply_to_can_message(buf, len, last_can_id );
  
  } //can signal processing over, if(Flag_Recv)
    else if( transmit_can_messages) {
     //say hi on the can bus
     if( (millis() - time_last_can_msg) > 4000 ){
       time_last_can_msg = millis();
       unsigned char newmsg[2] = { 32, 32 };
       unsigned char othermsg[0];
       CAN.sendMsgBuf( 0, 0, 0, othermsg);
       CAN.sendMsgBuf( 1154, 0, 2, newmsg);
       if( report_serial ) Serial.println("TX1B");
     }
    }
    
    int can_error_code = CAN.checkError();
 
    if( (can_error_code == 0) && (loop_calls % 20000 == 0) ){
      lcd.clear(); lcd.setCursor(0,0);
      lcd.print("CAN BUS ok.."); lcd.setCursor(0,1);
      lcd.print("Intensity: "); lcd.print(light_intensity);
    }
    else if( can_error_code  > 0) { //make sure there is no can error
    //if( CAN.checkError() > 0) { //make sure there is no can error

      
      if( report_serial ){Serial.println("CAN Error detected, reinitializing..");}
      lcd.clear(); lcd.setCursor(0,0); lcd.print("CAN error..");
      lcd.setCursor(0,1); lcd.print("reinit..");
      //delay(200);
      int init_counter = 1;
      while( CAN.begin(CAN_100KBPS)!=CAN_OK){
         if( init_counter == 10 ){
             lcd.clear(); lcd.setCursor(0,0); lcd.autoscroll(); 
             lcd.print("CAN REINIT PERMANENT FAIL. CHECK CONNECTIONS, PRESS RESET OR TRY AGAIN TOMORROW");
             while( 1 == 1 ){ delay(5000); }
         }
         init_counter++; 
      }
      set_tecan_can_filters();
      lcd.clear(); lcd.setCursor(0,0); lcd.print("CAN reinit OK");
    }
  
    
    
    if( digitalRead( button_pin ) == HIGH ){
     for(int led = 0; led < num_led_pins; ++led) analogWrite( led_pins[led], light_intensity);
     if( report_serial ) {Serial.print("button check, aread value is "); Serial.println(analogRead(int_pot_pin));}
     delay(500);
     for(int led = 0; led < num_led_pins; ++led) digitalWrite( led_pins[led], LOW);
    }
    
    int cur_aread = analogRead( int_pot_pin );
    if( abs( cur_aread - aread_value) > 5 ){
      aread_value = cur_aread;
      light_intensity = map( aread_value, 0, 1023, 0, 255 );
      lcd.clear(); lcd.setCursor(0,0);
      lcd.print("Intensity: "); lcd.print(light_intensity);
    }
    
}

/*********************************************************************************************************
  END FILE
*********************************************************************************************************/